package ru.cleptoman.flv {
	import adobe.utils.CustomActions;
	import flash.display.BitmapData;
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	import flash.events.TimerEvent;
	import flash.utils.ByteArray;
	import flash.utils.Timer;
	import ru.cleptoman.flv.audio.*;
	import ru.cleptoman.flv.events.FLVEncoderEvent;
	import ru.cleptoman.flv.video.*;
	import ru.cleptoman.flv.flvmetadata.FLVMetaData;
	import ru.cleptoman.flv.utils.BytesUtils;
	import ru.cleptoman.flv.utils.ScreenVideoPacket;
	
	/**
	 * ...
	 * @author Aleksey Kutov aka cleptoman
	 */
	public class FLVEncoder extends EventDispatcher {
		
		private const SUPPORT_AUDIO_CODECS:Vector.<uint>		= new <uint>[FLVAudioCodec.MP3,FLVAudioCodec.PCM_LITTLE_ENDIAN,FLVAudioCodec.PCM_PLATFORM_ENDIAN];
		private const QUEUE_EXEC_DELAY:uint						= 30;
		
		private var _destroyInputData:Boolean;
		private var _flvFile:ByteArray;
		private var _framerate:Number;
		private var _prevTagSize:uint;
		private static var _instance:FLVEncoder;
		private var _totalTags:Number;
		private var _currentTag:Number;
		private var _actionsQueue:Vector.<Action>;
		private var _queueExecTimer:Timer;
		
		public function FLVEncoder() {
			super();
			if (_instance) {
				throw new Error("FLVEncoder is singleton");
			}else{
				_instance					= this;
			}
		}

		public function encode(videoTags:FLVVideoTags, audioTags:FLVAudioTags = null, destroyInputData:Boolean = true):void {
			_destroyInputData				= destroyInputData;
			_flvFile						= new ByteArray();
			_currentTag						= 0;
			_prevTagSize					= 0;
			_framerate						= videoTags.config.framerate;
			var duration:Number				= videoTags.tags.length / _framerate * 1000;
			
			if (_queueExecTimer) {
				_queueExecTimer.stop();
			}else {
				_queueExecTimer				= new Timer(QUEUE_EXEC_DELAY);
				_queueExecTimer.addEventListener(TimerEvent.TIMER, onTickExecTimer);
			}
			
			_actionsQueue					= new Vector.<Action>;
			
			_totalTags						= videoTags.tags.length;
			_totalTags						+= (audioTags && audioTags.tags) ? audioTags.tags.length : 0;
			
			/**
			 * формируем загаловок
			 */
			doHeader(Boolean(audioTags && audioTags.tags && audioTags.tags.length));
			
			/**
			 * пишем мету
			 */
			doMetaData(duration, videoTags.config, audioTags ? audioTags.config : null);
			/**
			 * пишем аудио, если есть
			 */
			if(audioTags){
				doAudio(audioTags);
			}
			/**
			 * пишем видео
			 */
			doVideo(videoTags);
			
			/**
			 * запускаем таймер
			 */
			_queueExecTimer.start();

		}
		
		private function onTickExecTimer(e:TimerEvent):void {
			var action:Action		= _actionsQueue[_currentTag];
			_currentTag++;
			var ev:FLVEncoderEvent;
			if (_currentTag === _totalTags) {
				_queueExecTimer.stop();
				_queueExecTimer.removeEventListener(TimerEvent.TIMER, onTickExecTimer);
				_queueExecTimer			= null;
				ev						= new FLVEncoderEvent(FLVEncoderEvent.ENCODE_COMPLETE, 1);
			}else {
				ev						= new FLVEncoderEvent(FLVEncoderEvent.ENCODE_PROGRESS, Number(_currentTag / _totalTags));
				action.method.apply(this, action.params);
			}
			super.dispatchEvent(ev);
		}
		
		private function doVideo(videoTags:FLVVideoTags):void {
			
			var calculator:Object;
			switch(videoTags.config.codecID) {
				case FLVVideoCodec.SCREEN_VIDEO:
				calculator					= ScreenVideoPacket;	
				break;
				default:
				throw new Error(String(videoTags.config.codecID) + " video codec id is not supported in this version");
			}
			var tags:Vector.<FLVVideoTag> = videoTags.tags;
			var width:Number				= videoTags.config.videoWidth;
			var height:Number				= videoTags.config.videoHeight;
			

			var currTagIndex:uint;
			var totalTagIndex:uint			= tags.length;
			
			
			var currentTag:FLVVideoTag;
			var action:Action;
			
			
			for (currTagIndex = 0; currTagIndex < totalTagIndex ; currTagIndex++ ) {
				currentTag					= tags[currTagIndex];
				action						= new Action();
				action.method				= execVideoTag;
				action.params				= [currentTag, width, height, calculator.calculate];
				_actionsQueue.push(action);
			}
			
		}

		private function doAudio(audioTags:FLVAudioTags):void {
			
			var config:FLVAudioConfig		= audioTags.config;
			
			var isSupportedCodec:Boolean	= (SUPPORT_AUDIO_CODECS.indexOf(config.codecID) !== -1);
			if(isSupportedCodec){
				
				var tags:Vector.<FLVAudioTag> = audioTags.tags;
				
				var numTags:uint			= tags.length;
				
				var audioTag:FLVAudioTag;
				var currTagIndex:uint;
				
				var action:Action;
				
				var headerByte:Number = config.codecID << 4;
				headerByte |= config.soundRate << 2;
				headerByte |= 1 << 1;
				headerByte |= config.soundType << 1;
				
				for (currTagIndex = 0; currTagIndex < numTags; currTagIndex++ ) {

					audioTag					= tags[currTagIndex];
					action						= new Action();
					action.method				= execAudioTag;
					action.params				= [audioTag,headerByte];
					_actionsQueue.push(action);
				
				}
				
			}else {
				throw new Error(String(config.codecID) + " audio codec id is not supported in this version");
			}
			
		}
		
		private function doMetaData(duration:Number,videoConfig:FLVVideoConfig, audioConfig:FLVAudioConfig = null):void {
			var meta:ByteArray		= new ByteArray();
			
			meta.writeByte(18);

			var data:FLVMetaData 	= new FLVMetaData();
			data.duration			= duration || 0;
			data.width				= videoConfig.videoWidth || 0;
			data.height				= videoConfig.videoHeight || 0;
			data.framerate			= videoConfig.framerate || 0;
			data.videocodecid		= videoConfig.codecID;
			if (audioConfig) {
				data.audiocodecid	= audioConfig.codecID;
			}
			data.creationdate		= new Date().toString();
			var dataBytes:ByteArray = data.getBytes();
			
			BytesUtils.writeUnsignedInt24(meta, dataBytes.length);
			BytesUtils.writeUnsignedInt24(meta, 0);
			meta.writeByte(0);
			BytesUtils.writeUnsignedInt24(meta, 0);
	
			meta.writeBytes( dataBytes );
			
			_prevTagSize 			= meta.length;
			_flvFile.writeBytes(meta);
		}
		
		private function doHeader(hasAudio:Boolean):void {
			var header:ByteArray = new ByteArray();
			header.writeByte(0x46);
			header.writeByte(0x4C);
			header.writeByte(0x56);
			header.writeByte(0x01);
			header.writeByte(hasAudio ? 5 : 4); 
			header.writeUnsignedInt(0x09);
			_flvFile.writeBytes(header);
			/**
			 * перед любым тэгом нужно записать размер предыдущего тэга. в первом случае нужно записать ноль
			 */
			_flvFile.writeUnsignedInt(_prevTagSize);
		}
		
		private function execAudioTag(audioTag:FLVAudioTag, headerByte:uint):void {
			var audio:ByteArray			= new ByteArray();
			var tag:ByteArray;
			tag						= new ByteArray();
			audio.writeUnsignedInt(_prevTagSize);
			tag.writeByte( 0x08 ); 
			BytesUtils.writeUnsignedInt24(tag, audioTag.sample.length + 1); 
			BytesUtils.writeUnsignedInt24(tag, audioTag.timeStamp);
			tag.writeByte(0);	
			BytesUtils.writeUnsignedInt24(tag, 0);
			tag.writeByte(headerByte);
			tag.writeBytes(audioTag.sample);
			audio.writeBytes( tag );
			_flvFile.writeBytes(audio);
			if(_destroyInputData){
				audioTag.sample		= null;
			}
			_prevTagSize			= tag.length;
		}
		
		private function execVideoTag(currentTag:FLVVideoTag, width:Number, height:Number, calcMethod:Function):void {
			var video:ByteArray				= new ByteArray();
			var tag:ByteArray;
			tag							= new ByteArray();
			video.writeUnsignedInt(_prevTagSize);
			var videoTagData:ByteArray;
			videoTagData				= calcMethod(currentTag.frameType, 32, width, 32, height, currentTag.video);
			tag.writeByte( 0x09 );
			BytesUtils.writeUnsignedInt24(tag, videoTagData.length);
			BytesUtils.writeUnsignedInt24(tag, currentTag.timeStamp);
			tag.writeByte(0);
			BytesUtils.writeUnsignedInt24(tag, 0);
			tag.writeBytes( videoTagData );
			_prevTagSize = tag.length;
			video.writeBytes(tag);
			if(_destroyInputData){
				currentTag.video.dispose();
				currentTag.video		= null;
			}
			_flvFile.writeBytes(video);
		}
		
		public function get flvFile():ByteArray {
			return _flvFile;
		}
		
		public static function getInstance():FLVEncoder {
			return new FLVEncoder();
		}
		
	}

}

class Action {
	private var _method:Function;
	private var _params:Array;
	
	public function get method():Function {
		return _method;
	}
	
	public function set method(value:Function):void {
		_method = value;
	}
	
	public function get params():Array {
		return _params;
	}
	
	public function set params(value:Array):void {
		_params = value;
	}
}
